import inspect
from functools import lru_cache
from typing import Any, Callable, Dict, List, Optional, TypeVar, Union, overload
from typing_extensions import Literal, get_args


@lru_cache(maxsize=250)
def get_func_args(func: Callable[[Any], Any]) -> List[str]:
    """Returns a list of arguments for the function"""

    sig = inspect.signature(func)

    return [
        arg_name
        for arg_name, param in sig.parameters.items()
        if param.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD
    ]


@overload
def get_specialized_type_var_map(
    cls: type,
    *,
    include_type_vars: Literal[True],
) -> Optional[Dict[TypeVar, Union[TypeVar, type]]]:
    ...


@overload
def get_specialized_type_var_map(
    cls: type,
    *,
    include_type_vars: Literal[False] = ...,
) -> Optional[Dict[TypeVar, type]]:
    ...


@overload
def get_specialized_type_var_map(
    cls: type,
    *,
    include_type_vars: bool,
) -> Optional[
    Union[Optional[Dict[TypeVar, type]], Dict[TypeVar, Union[TypeVar, type]]]
]:
    ...


def get_specialized_type_var_map(cls: type, *, include_type_vars: bool = False):
    """Get a type var map for specialized types.

    Consider the following:

        >>> class Foo(Generic[T]):
        ...     ...
        ...
        >>> class Bar(Generic[K]):
        ...     ...
        ...
        >>> class IntBar(Bar[int]):
        ...     ...
        ...
        >>> class IntBarSubclass(IntBar):
        ...     ...
        ...
        >>> class IntBarFoo(IntBar, Foo[str]):
        ...     ...
        ...

    This would return:

        >>> get_specialized_type_var_map(object)
        None
        >>> get_specialized_type_var_map(Foo)
        {}
        >>> get_specialized_type_var_map(Foo, include_type_vars=True)
        {~T: ~T}
        >>> get_specialized_type_var_map(Bar)
        {~T: ~T}
        >>> get_specialized_type_var_map(IntBar)
        {~T: int}
        >>> get_specialized_type_var_map(IntBarSubclass)
        {~T: int}
        >>> get_specialized_type_var_map(IntBarFoo)
        {~T: int, ~K: str}

    """
    orig_bases = getattr(cls, "__orig_bases__", None)
    if orig_bases is None:
        # Not a specialized type
        return None

    type_var_map = {}

    for base in orig_bases:
        # Recursively get type var map from base classes
        base_type_var_map = get_specialized_type_var_map(base)
        if base_type_var_map is not None:
            type_var_map.update(base_type_var_map)

        args = get_args(base)
        origin = getattr(base, "__origin__", None)

        params = origin and getattr(origin, "__parameters__", None)
        if params is None:
            params = getattr(base, "__parameters__", None)

        if not params:
            continue

        type_var_map.update(
            {
                p: a
                for p, a in zip(params, args)
                if include_type_vars or not isinstance(a, TypeVar)
            }
        )

    return type_var_map



from .context import StrawberryChannelsContext
from .handlers.graphql_transport_ws_handler import GraphQLTransportWSHandler
from .handlers.graphql_ws_handler import GraphQLWSHandler
from .handlers.http_handler import GraphQLHTTPConsumer
from .handlers.ws_handler import GraphQLWSConsumer
from .router import GraphQLProtocolTypeRouter

__all__ = [
    "GraphQLProtocolTypeRouter",
    "GraphQLWSHandler",
    "GraphQLTransportWSHandler",
    "GraphQLHTTPConsumer",
    "GraphQLWSConsumer",
    "StrawberryChannelsContext",
]



from __future__ import annotations

from inspect import getframeinfo, stack
from pathlib import Path
from typing import TYPE_CHECKING, Optional, Type

from strawberry.exceptions.utils.source_finder import SourceFinder
from strawberry.utils.cached_property import cached_property

from .exception import StrawberryException

if TYPE_CHECKING:
    from strawberry.union import StrawberryUnion

    from .exception_source import ExceptionSource


class InvalidUnionTypeError(StrawberryException):
    """The union is constructed with an invalid type"""

    invalid_type: object

    def __init__(self, union_name: str, invalid_type: object) -> None:
        from strawberry.custom_scalar import ScalarWrapper

        self.union_name = union_name
        self.invalid_type = invalid_type

        # assuming that the exception happens two stack frames above the current one.
        # one is our code checking for invalid types, the other is the caller
        self.frame = getframeinfo(stack()[2][0])

        if isinstance(invalid_type, ScalarWrapper):
            type_name = invalid_type.wrap.__name__
        else:
            try:
                type_name = invalid_type.__name__  # type: ignore
            except AttributeError:
                # might be StrawberryList instance
                type_name = invalid_type.__class__.__name__

        self.message = f"Type `{type_name}` cannot be used in a GraphQL Union"
        self.rich_message = (
            f"Type `[underline]{type_name}[/]` cannot be used in a GraphQL Union"
        )
        self.suggestion = (
            "To fix this error you should replace the type a strawberry.type"
        )
        self.annotation_message = "invalid type here"

    @cached_property
    def exception_source(self) -> Optional[ExceptionSource]:
        path = Path(self.frame.filename)

        source_finder = SourceFinder()

        return source_finder.find_union_call(path, self.union_name, self.invalid_type)


class InvalidTypeForUnionMergeError(StrawberryException):
    """A specialized version of InvalidUnionTypeError for when trying
    to merge unions using the pipe operator."""

    invalid_type: Type

    def __init__(self, union: StrawberryUnion, other: object) -> None:
        self.union = union
        self.other = other

        # assuming that the exception happens two stack frames above the current one.
        # one is our code checking for invalid types, the other is the caller
        self.frame = getframeinfo(stack()[2][0])

        other_name = getattr(other, "__name__", str(other))

        self.message = f"`{other_name}` cannot be used when merging GraphQL Unions"
        self.rich_message = (
            f"`[underline]{other_name}[/]` cannot be used when merging GraphQL Unions"
        )
        self.suggestion = ""
        self.annotation_message = "invalid type here"

    @cached_property
    def exception_source(self) -> Optional[ExceptionSource]:
        source_finder = SourceFinder()

        return source_finder.find_union_merge(self.union, self.other, frame=self.frame)
